'''
  $Id: __main__.py 1473 2019-05-10 10:01:09Z sschaiter $

  Jira Helper CLI

  (c) 2019 ITEG GmbH http://www.iteg.at

'''

# The bootstrapping main for setting up logging before import the actual main program.

from clazzes.jiraclient.main import cliMain, ISSUETYPES, PRIORITIES
from argparse import ArgumentParser
import sys
import logging

log = logging.getLogger(__name__)

 

class JiraClientParser(ArgumentParser):
    #Structure inspired by https://chase-seibert.github.io/blog/2014/03/21/python-multilevel-argparse.html#
    
    def __init__(self):
        """
        The top-level parser which only parses the subcommand and directs it to the appropriate subparser, if possible.
        """
        ArgumentParser.__init__(self)
        self.description = 'Command Line Interface Tool for Jira.'
        self.usage =  '''jiraclient <subcommand> <args>
        
List of subcommands:

listissues - get a comprehensive list of issues found for a given project.
createoauth - request a new OAuth Token.
createissue - create a new Jira issue.
initconfig - Initialize a jiraclient config file.

Run -h with one of the subcomands to get usage for it.'''
        self.add_argument('command', help="Subcommand to parse.", choices = ["listissues", "createoauth", "createissue", "initconfig"])

        #Only parse command in order to call appropriate subparser afterwards with rest of arguments

    def parse(self):
        parsedCmd = self.parse_args(sys.argv[1:2]).command
        if not hasattr(self, parsedCmd):
            print('Unrecognized command')
            self.print_help()
            exit(64)
        #Retrieve function object representing subparser from within this classes' scope and parse it
        #Returns the parsed variables
        return getattr(self, parsedCmd)()
        
    def addUniversalArgs(self):
        """
        Add the ubiquitous arguments used for every subcommand to the argParser.
        """
        self.add_argument("-c", "--config", help = "Specify the path to the config file to read configuration from.")
        self.add_argument("-u", "--username", help = "The username to authenticate with.")
        self.add_argument("-o", "--oauthtoken", help = "The OAuth token to authenticate with.")
        self.add_argument("-p", "--password", help = "The password to authenticate with.")
        self.add_argument("-d", "--debug", help = "Enable debug mode.", action="store_true")
        self.add_argument("-a", "--auth", help = "The authentication method to use.", 
                            choices = ["OAUTH", "BASICAUTH"] )
        self.add_argument('-b', "--baseurl", help = "The base URL of the jira server. E.g. https://example.iteg.com/ ")
        
    def listissues(self):
        self.description = "Print a list of issues of a given project name, if it is found."
        self.usage = "jira-client listissues <projectName>"
        self.add_argument("project", help = "The name of the project to search for.")
        parsedArgs = vars(self.parse_args(sys.argv[1:]))
        return parsedArgs
        
    def initconfig(self):
        self.description = "Initialize or supplement a config file."
        self.usage = """jira-client initconfig [--conffile = <path to custom config file>] [<parameters to be written to config>]
Parameters that can be specified are:
--username, --password, --auth, --oauthtoken, --oauthsecrettoken, --pkeyloc, --consumerkey, --consumersecret  
Non-specified parameters in this list will be asked to be manually entered(or ignored and left unchanged).
If --conffile is not specified, JiraClient will ask from the use to input filepath manually.
Can create files not already existing.
        """
        self.add_argument("--conffile", help = "Specifiy a path to write the values to. If it isn't already a config file, \
        try to create new")
        self.add_argument("-k", "--consumerkey", help = "The name of the OAuth public key as configured on the Jira Server.")
        self.add_argument("-s", "--consumersecret", help = "The OAuth secret as configured on the Jira Server.")
        self.add_argument("--pkeyloc", help = "The file path to the RSA private key needed for OAuth signatures")
        self.add_argument("--oauthsecrettoken", help = "The OAuth Secret Token to the specified OAuth Token")
        parsedArgs = vars(self.parse_args(sys.argv[1:]))
        return parsedArgs
        
    def createissue(self):
        self.description = "Create a new jira issue."
        #TODO Implement createIssue
        self.usage = """jira-client createissue --project <project name> --summary <summary/headline> --desc <description> --issuetype <issuetype> --priority <priority>
[--assignee <assignee user name>] [--parent <parent issue id>]
Use '-' for 'desc' parameter to use piped input as description.
assignee is optional, parent is only needed when issue type is 'subtask'
        """
        self.add_argument("--project", help = "The name of the project to add the issue to.")
        self.add_argument("--summary", help = "The issue's title in Jira.")
        self.add_argument("--desc", help = "More detailed description. Specify '-'(single hyphen) as the value to use piped input instead.")
        self.add_argument("--issuetype", help = "The type of issue.", choices = ISSUETYPES, type = str.lower)
        self.add_argument("--priority", help = "The issue's requested priority. Numbers 1-5 stand for %s respectively." % str(PRIORITIES), choices = [1,2,3,4,5], type = int )
        self.add_argument("--assignee", help = "The user assigned to solve the issue. ")
        self.add_argument("--parent", help = "ID of the parent issue. Only needed/used when creating a subtask.")
        parsedArgs = vars(self.parse_args(sys.argv[1:]))
        return parsedArgs
    
    def createoauth(self):
        self.description = "Request a new OAuth Token to authorize with."
        self.usage = "jira-client createoauth [--consumerkey <consumer key>] [--consumersecret <consumer secret>]"
        self.add_argument("-k", "--consumerkey", help = "The name of the OAuth public key as configured on the Jira Server.")
        self.add_argument("-s", "--consumersecret", help = "The OAuth secret as configured on the Jira Server.")
        self.add_argument("--pkeyloc", help = "The file path to the RSA private key needed for OAuth signatures")
        parsedArgs = vars(self.parse_args(sys.argv[1:]))
        return parsedArgs

def main():
    jcp = JiraClientParser()
    jcp.addUniversalArgs()
    parsedArgs = jcp.parse()
    try:
        cliMain(**parsedArgs)
    except KeyboardInterrupt:
        print("Caught Keyboard Interrupt. Exiting Jira-Client...")
        exit(64)
    except Exception as e:
        print("%s occured" % type(e).__name__)
        if hasattr(e, "args") and e.args is not None:
            print(e.args[0])
        if hasattr(e, 'message'):
            print(e.message)
        if hasattr(e, 'msg'):
            print(e.msg)
        print("Exiting Jira-Client...")
    
if __name__ == "__main__":
    main()
