'''
Created on 21.05.2019

@author: sschaiter
'''

import logging
import oauth2 as oauth #oauth2 library is actually the 2nd revision of an OAuth 1.0a library
from urllib import parse as urlparse
from clazzes.jiraclient.main import CONFIGKEYS
from clazzes.jiraclient.utils import setConfigParams, sendJqlQuery, \
makeApiCall, getProjectIssueTypeIds, makeConfigParser, createConfigFile
from sys import exit
from datetime import datetime as dt

log = logging.getLogger(__name__)


def getJiraIssues(projectName : str = None, authMethod = 'OAUTH', oAuthHeader : str = None, username : str = None, 
                  password : str = None):
    """
    Obtain the list of issues for a given project name.
    
    :param projectName: The project's descriptor. Corresponds to "key" parameter of the project JSON on Jira.
    :param authMethod: Which authentication method to use.
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: username to authenticate with. Only needed if authMethod is 'BASICAUTH'.
    :param password: password to authenticate with. Only needed if authMethod is 'BASICAUTH'
    :return: A dictionary holding all issues belonging to the project searched for. Keys are the issue names, 
    values the entire JSON object.
    """

    projectJson = getProjectByName(projectName, authMethod, oAuthHeader, username, password)
    log.debug('Retrieved list of issues for project %s' % projectName)
    issueList = projectJson['issues']
    if not issueList:
        log.exception('Retrieved list of issues for project %s was empty.' % projectName)
        raise ValueError('Retrieved list of issues for project %s was empty.' % projectName)
    issueDict = {}
    for issue in issueList:
        issueDict[issue['key']] = issue
    return issueDict

def listJiraIssues(projectName, authMethod = 'OAUTH', oAuthHeader : str = None, 
                   username : str = None, password : str = None):
    """
    Print to output an exhaustive list of issues belonging to the given project.
    
    :param projectName: The project's descriptor. Corresponds to "key" parameter of the project JSON on Jira.
    :param authMethod: Which authentication method to use.
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: username to authenticate with. Only needed if authMethod is 'BASICAUTH'.
    :param password: password to authenticate with. Only needed if authMethod is 'BASICAUTH'
    """
    jiraIssues = getJiraIssues(projectName, authMethod, oAuthHeader, username, password)
    
    issueStrings = []
    for issueData in jiraIssues.values():
        name = issueData.get('key')
        fields = issueData.get('fields')
        
        category = fields.get('issuetype').get('name')
        creator = fields.get('creator').get('displayName')
        created = fields.get('created')
        created = dt.strptime(created, "%Y-%m-%dT%H:%M:%S.000%z").strftime("%a, %d.%m.%Y %H:%M:%S (%Z)")
        priority = fields.get('priority').get('id') + ' - ' + fields.get('priority').get('name')
        assignee = fields.get('assignee').get('displayName') if fields.get('assignee') is not None else "Nobody"
        if fields.get('updated', None) is not None:
            lastUpdated = fields.get('updated')
            lastUpdated = dt.strptime(lastUpdated, "%Y-%m-%dT%H:%M:%S.000%z").strftime("%a, %d.%m.%Y %H:%M:%S (%Z)")
        else:
            lastUpdated = 'never'
        description = fields.get('description') if fields.get('description', None) is not None else 'No description'
        
        issueString = name + ' (' + category + "):\n" \
        + "Priority " + priority + '\n' \
        + description + '\n' \
        + 'Created by ' + creator + ' at ' + created + '\n' \
        + 'Assigned to ' + assignee + '\n' \
        + 'Last updated: ' + lastUpdated + '\n----------------------------'
        
        issueStrings.append(issueString)
    
    for issueString in issueStrings:
        print(issueString)

def getProjectByName(projectName, authMethod, oAuthHeader, username, password):    
    """
    Get the JSON summary of a project by the project name. 
    
    :param projectName: The project name. It corresponds to the value to the "key" key in the project JSON.
    :param authMethod: Which authentication method to use.
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: username to authenticate with. Only needed if authMethod is 'BASICAUTH'.
    :param password: password to authenticate with. Only needed if authMethod is 'BASICAUTH'
    """    
    
    from clazzes.jiraclient.main import jiraApiUrl
    metadata = makeApiCall(jiraApiUrl + "issue/createmeta", authMethod, oAuthHeader, username, password, 
                           params = {'projectKeys' : projectName})
    try:
        projectId = metadata['projects'][0]['id'] 
    except IndexError:
        raise IndexError("Did not find metadata for project %s. It might be a spelling error, or the authenticated user has no browsing rights." % projectName )
    projectJson = sendJqlQuery({"project" : projectId}, authMethod, oAuthHeader, username, password)
    return projectJson

def createOAuthToken(consumerKey, consumerSecret, sigMethod):
    """
    Initialize the process of creating a new OAuth Access Token and its corresponding Secret Token. Requires manual authorization 
    by the user inside an internet browser since OAuth1 does not support automatic authorization.
    OAuth Token created uses OAuth 1.0a mechanism as Jira does not support OAuth 2. Authorization happens by configuring
    the client on a sufficiently able account on the Jira Server. This also includes setting a public key for the client.
    A valid signature done with the private key proves identity to the Jira Server.
    Please note that the all requests to jira with the Access Token also need a valid signature and the appropriate headers.
    Those are done by the requests.oauthlib
    See https://developer.atlassian.com/server/jira/platform/oauth/ for more details.
     
    :param consumerKey: How the public key is identified on the Jira Server.
    :param consumerSecret: A (randomly generated) secret that is also saved on the Jira Server.
    :param sigMethod:  The Signature Method to be used. Must be instance of a class that has the "name" parameter set to a 
    valid value(ex. "RSA-SHA1") as well as implement the sign() method. Required for the authentication signature needed to
    prove ownership of the corresponding Private Key to the Server.
    """
    
    from clazzes.jiraclient.main import requestTokenUrl, authorizeUrl, accessTokenUrl
    
    consumer = oauth.Consumer(consumerKey, consumerSecret)
    client = oauth.Client(consumer)
    client.set_signature_method(sigMethod)

    resp, content = client.request(requestTokenUrl, "POST")
    if resp['status'] != '200':
        raise Exception("Invalid response %s: %s" % (resp['status'], content))
    requestToken = dict(urlparse.parse_qsl(content.decode('UTF-8')))
    log.debug("Retrieved OAuth request token %s with secret %s" % 
             (requestToken['oauth_token'], requestToken['oauth_token_secret']))
    #Manual authorization:
    print("Request token recieved, please authorize me in the Browser via the following URL:")
    print("%s?oauth_token=%s\n" % (authorizeUrl, requestToken['oauth_token']))
    
    authorized = 'n'
    while authorized.lower() == 'n':
        authorized = input('Have you authorized me? (Y/n) ')
        if not authorized: #just pressing enter defaults to "Yes"
            authorized = 'Y'
        if authorized.lower() not in ("y", "n"):
            authorized = 'n'
            
    token = oauth.Token(requestToken['oauth_token'], 
                        requestToken['oauth_token_secret'])
    client = oauth.Client(consumer, token)
    client.set_signature_method(sigMethod)
    
    resp, content = client.request(accessTokenUrl, "POST")
    accessToken = dict(urlparse.parse_qsl(content.decode('UTF-8')))
    log.debug('Retrieved access token was %s' % (accessToken))
    
    gotValidAnswer = False
    while not gotValidAnswer:
        doSave = input("Save the Token to a config file? (Y/n) ")
        if doSave:
            if doSave.lower() in ('y', 'n'):
                gotValidAnswer = True
        else:
            doSave = 'y'
            gotValidAnswer = True
    if doSave == 'y':
        configPath = createConfigFile()
        setConfigParams({'oauthtoken' : accessToken['oauth_token'], 'oauthsecrettoken' : accessToken['oauth_token_secret']}, 
                             configPath)
        
def initializeConfigFile(configPath: str = None, configParser = None, **confParams):
    """
    Convenience command that allows setting of config parameters in the CLI.
    When using this command, supplying any options that also be read from configs will automatically write them to the config.
    Using the "-c" option for config with this command will assume that values should be written to that config file.
    If no config file was submitted, it will ask for a path or if it should write to user or global config file.  
    If any parameter was not submitted, it will ask what value to write for it, which can be ignored to not change anything.
    
    :param configPath: Path to the config file to be modified. This can be set by submitting a value for the -c option.
    :param configParser: Submit a :class:~`ConfigParser` instance to be used. If None, one is created.
    :param confParams: Optional parameters that must be part of allowed config params. Any that are not set will be asked for manually.
    """
    
    for key in confParams.keys():
        if key not in CONFIGKEYS:
            raise ValueError("Submitted parameter %s not eligible to be set in config files." % key)
    
    if not configPath:
        configPath = createConfigFile(None, True)
    else:
        configPath = createConfigFile(configPath, False)
            
    if not configPath:
        #This means that user declined to create a config file
        print("Creation of config file was declined. Exiting...")
        exit(64)        
            
    if configParser == None:
        configParser = makeConfigParser()
    if not configParser.read(configPath):
        raise ValueError("Could not parse config file at path %s" % configPath)
        
    if not configParser.has_section('CLI'):
        configParser.add_section('CLI')
        
    remainingKeys = CONFIGKEYS
    for key, value in confParams.items():
        configParser.set("CLI", key.lower(), value)
        remainingKeys.remove(key)
    if remainingKeys:
        print("Enter remaining configurable values. Leave empty to not change that value:\n")
    for key in remainingKeys:
        value = input(key + ": ")
        if value.strip() == "":
            continue
        configParser.set("CLI", key.lower(), value)
    try:
        with open(configPath, 'w') as fh:
            configParser.write(fh)
    except IOError:
        raise IOError("Jira-Client does not have permissions to write to config file %s" % configPath)
    log.info("Written to config file at path %s" % configPath)
        
def createIssue(projectName, issueSummary, issueDesc, issueType, priority, authMethod, oAuthHeader, username, password, 
                assignee = None, parent = None):
    
    """
    Create an issue on Jira with given parameters.
    
    :param projectName: The name of the project to put the issue onto. In the project description JSON, this is found under
    the key 'key'.
    :param issueSummary: The headline of the issue. Should depict the issue's core problem/task in one sentence.
    :param issueDesc: This should elaborate on the problem/task more extensively.
    :param issueType: The identifier of the issue type to give to the issue. Usually ranges from 5-Trivial to 1-Blocker.
    :param authMethod: Which authentication method to use. Valid ones are BASICAUTH and OAUTH. 
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: username to authenticate with. Only needed if authMethod is 'BASICAUTH'.
    :param password: password to authenticate with. Only needed if authMethod is 'BASICAUTH'
    """
    
    from clazzes.jiraclient.main import jiraApiUrl, baseUrl
    
    projectJson = getProjectByName(projectName, authMethod, oAuthHeader, username, password)
    if not projectJson:
        raise ValueError("Received empty JSON when asking for metadata of project with name %s" %projectName)
    
    issueTypeId = getProjectIssueTypeIds(projectName, authMethod, oAuthHeader, username, password).get(issueType, None)
    if not issueTypeId:
        raise ValueError('issueType %s is not part of the issue type list of project %s' % (issueType, projectName) )

    
    issueJson = {
        "fields" : {
            "project" : 
            {
                "key" : projectName
            },
            "summary" : issueSummary,
            "description" : issueDesc,
            "issuetype" : {
                "id" : issueTypeId
            },
            "priority" : {
                "id" : str(priority)
            },
            "assignee" : {
                "name" : assignee
            },
            "parent" : {
                "id" : parent
            }
        }
    }


    respJson = makeApiCall(jiraApiUrl + "issue/", authMethod, oAuthHeader, username, password, json = issueJson, method = "POST")
    newIssueId = respJson["id"]

    log.info("New Issue successfully created.(ID %s, name %s" % (newIssueId, projectName))

    print(baseUrl + 'browse/' + newIssueId)