'''
  $Id: main.py 1504 2019-05-10 10:01:09Z sscha $

  Jira Helper CLI

  (c) 2019 ITEG GmbH http://www.iteg.at

'''

import os
import logging
from select import select
from requests_oauthlib import OAuth1
from sys import stdin

ETC_CONFIG_FILEPATH = "/etc/jiraclient/jiraclient.conf"
USER_CONFIG_FILEPATH = os.path.expanduser("~/.jiraclient/jiraclient.conf")

privateKeyLocation = None
baseUrl = None
oAuthBaseUrl = None
requestTokenUrl = None
accessTokenUrl = None
authorizeUrl = None
jiraApiUrl = None
jiraAuthUrl = None



#Different kinds of issues in Jira. Ordered by their IDs in ascending order, with index 0("bug") corresponding to id 1 on Jira.
ISSUETYPES = ["bug", "feature", "task", "improvement", "subtask", "epic", "story", "technical" ]
ISSUETYPES_TRANSLATION = {"bug" : "Bug", "feature" : "Neue Funktion", "task" : "Aufgabe", "improvement" : "Verbesserung",
                         "subtask" : "Unteraufgabe", "epic" : "Epic", "story" : "Story", "technical" : "Technische Aufgabe"} 

#Different priorities in Jira. Ordered by their IDs in ascending order, with index 0("Blocker") corresponding to id 1 on Jira.
PRIORITIES = ['Blocker', 'Kritisch', 'Schwer', 'Unwesentlich', 'Trivial']

#List of keys that have can have values set for them in the config files
CONFIGKEYS = ["username", "password", "oAuthToken", "consumerKey", "consumerSecret", 
              "oAuthSecretToken", "auth", "privateKeyLocation", "baseurl"]

REQUIREDVARS = {
    'initconfig' : [],
    'listissues' : ['baseUrl', 'auth', 'username', 'password', 'oAuthToken', 'oAuthSecretToken', 'privateKeyLocation', 'consumerKey', 'consumerSecret', 'project'],
    'createissue' : ['baseUrl', 'auth', 'username', 'password', 'oAuthToken', 'oAuthSecretToken', 'privateKeyLocation', 'consumerKey', 'consumerSecret', 'issueSummary', \
                     'desc', 'priority', 'issueType', 'project'],
    'createoauth' : ['baseUrl', 'privateKeyLocation', 'consumerKey', 'consumerSecret']
}
from clazzes.jiraclient.cmds import listJiraIssues, createOAuthToken,\
    initializeConfigFile, createIssue
from clazzes.jiraclient.utils import SignatureMethod_RSA_SHA1, makeConfigParser

log = logging.getLogger(__name__)


def cliMain(command=None, **args):
    
    log.debug("Running jiraclient %s with args %s" % (command, ["%s=%s" % (key, value) for (key, value) in args.items()] ) )
    
    configParser = None
    if args.get('config', None) is None: #No custom config was specified, so try to find global and user configs
        
        if os.path.exists(ETC_CONFIG_FILEPATH):
            configParser = makeConfigParser()
            log.info("Reading global configParser file [%s]." % ETC_CONFIG_FILEPATH)
            configParser.read(ETC_CONFIG_FILEPATH, "utf-8")
        else:
            log.info("Skipping non-existent configParser file [%s]." % ETC_CONFIG_FILEPATH)        
            
        if os.path.exists(USER_CONFIG_FILEPATH):
            if not configParser:
                configParser = makeConfigParser()
            log.info("Reading user configParser file [%s]." % USER_CONFIG_FILEPATH)
            configParser.read(USER_CONFIG_FILEPATH, "utf-8")
        else:
            log.info("Skipping non-existent configParser file [%s]." % USER_CONFIG_FILEPATH)
        
    else: #Custom config was specified, use exclusively values from there  
        config = args['config']
        if config[0] == '~': #User path given
            config = os.path.expanduser(config)
        if os.path.exists(config):
            configParser = makeConfigParser()
            log.info("Reading custom configParser file [%s]." % config)
            configParser.read(config, "utf-8")
        else:
            raise ValueError("Specified config file cannot be found")
    
    baseurl = None
    username = None
    password = None
    oAuthToken = None
    oAuthSecretToken = None
    oAuthFile = None
    consumerKey = None
    consumerSecret = None
    auth = None
    project = None
    targetPath = None
    desc = None
    assignee = None
    parent = None
    cmdLineParams = [] #This is required for the initializeConfigFile() function to only take parameters submitted manually
    
    if configParser is not None and configParser.has_section("CLI"):
        
        username = configParser['CLI'].get('username')
        password = configParser['CLI'].get('password')
        oAuthToken = configParser['CLI'].get('oauthtoken')
        oAuthSecretToken = configParser['CLI'].get('oauthsecrettoken')
        consumerKey = configParser['CLI'].get('consumerkey')
        consumerSecret = configParser['CLI'].get('consumersecret')
        auth = configParser['CLI'].get('auth')
        privateKeyLocation = configParser['CLI'].get('privatekeylocation')
        baseUrl = configParser['CLI'].get('baseurl')
    
    notNoneArgs = {key : value for key, value in args.items() if value is not None}
    for key, value in notNoneArgs.items():
        if key == "username":
            username = value
            cmdLineParams.append("username")
        elif key == "password":
            password = value
            cmdLineParams.append("password")
        elif key == "oauthtoken":
            oAuthToken = value
            cmdLineParams.append("oAuthToken")
        elif key == "oauthsecrettoken":
            oAuthSecretToken = value
            cmdLineParams.append("oAuthSecretToken")
        elif key == "consumerkey":
            consumerKey = value
            cmdLineParams.append("consumerKey")
        elif key == "consumersecret":
            consumerSecret = value
            cmdLineParams.append("consumerSecret")
        elif key == "auth":
            auth = value
            cmdLineParams.append("auth")
        elif key == "project":
            project = value
        elif key == "conffile":
            targetPath = value
            if targetPath[0] == '~':
                targetPath = os.path.expanduser(targetPath)
        elif key == "summary":
            issueSummary = value
        elif key == "issuetype":
            issueType = ISSUETYPES_TRANSLATION[value]
        elif key == "desc":
            desc = value
            #Hyphen means take piped input as description
            if desc == '-':
                print('Attempting to read piped stdin(If this runs indefinitely, something went wrong with piping)...')
                desc = stdin.read()
                print('Stdin read successfully')
        elif key == "priority":
            priority = value
        elif key == "pkeyloc":
            privateKeyLocation = value
            cmdLineParams.append("privateKeyLocation")
        elif key == "assignee":
            assignee = value
        elif key == "parent":
            issueTypeEnglish = list(ISSUETYPES_TRANSLATION.keys())[list(ISSUETYPES_TRANSLATION.values()).index(issueType)]
            if issueType and issueTypeEnglish == "subtask":
                parent = value
        elif key == "baseurl":
            baseUrl = value
            cmdLineParams.append("baseurl")
         
    if 'privateKeyLocation' in REQUIREDVARS[command] and privateKeyLocation:
        #Check for existence of private key file
        if privateKeyLocation[0] == '~':
            privateKeyLocation = os.path.expanduser(privateKeyLocation)
        if not os.path.isfile(privateKeyLocation):
            raise FileNotFoundError("This command requires a valid private key, but the specified private key location does not contain a file.")
  
    if 'baseUrl' in REQUIREDVARS[command] and baseUrl:
        #Must have slash at the end
        if baseUrl[-1] != '/':
            baseUrl += '/'
        
    localVars = locals()
    missingVars = [var for var in REQUIREDVARS[command] if localVars.get(var, None) is None]
    if 'auth' not in missingVars and command != 'createoauth':
        if localVars['auth'] == 'OAUTH':
            for var in ['username', 'password']:
                if var in missingVars:
                    missingVars.remove(var)
        elif localVars['auth'] == 'BASICAUTH':
            for var in ['oAuthToken', 'oAuthSecretToken', 'privateKeyLocation', 'consumerKey', 'consumerSecret']:
                if var in missingVars:
                    missingVars.remove(var)
    if missingVars:
        raise ValueError("Not all required parameters were provided. Please specify the following parameters via config or command line: " + ''.join(v.lower()  + ', ' for v in missingVars)[:-2] )
    
    if locals().get('baseUrl', None):
        oAuthBaseUrl = baseUrl + 'plugins/servlet/oauth/'
        requestTokenUrl = oAuthBaseUrl + 'request-token'
        accessTokenUrl = oAuthBaseUrl + 'access-token' 
        authorizeUrl = oAuthBaseUrl + 'authorize'
        jiraApiUrl = baseUrl + "rest/api/latest/"
        jiraAuthUrl = baseUrl + "rest/auth/latest/"
           
    for globalVar in ['oAuthBaseUrl', 'requestTokenUrl', 'baseUrl', 'accessTokenUrl', 'authorizeUrl', 'jiraApiUrl', 'jiraAuthUrl', 'privateKeyLocation']:
        #This is unfortunately needed for the other modules that import these to have the correct values(instead of just None)
        if globalVar in locals():
            globals()[globalVar] = locals()[globalVar]
    if locals().get('privateKeyLocation'):
        globals()['privateKeyLocation'] = privateKeyLocation 
    
    if consumerKey and consumerSecret and oAuthToken and oAuthSecretToken and privateKeyLocation and auth == "OAUTH":
        with open(privateKeyLocation, 'r') as fh:
            privateKey = fh.read()
        oAuthHeader = OAuth1(consumerKey, client_secret = consumerSecret, resource_owner_key = oAuthToken, 
                             resource_owner_secret = oAuthSecretToken, signature_method = 'RSA-SHA1',
                             rsa_key = privateKey)
    else:
        oAuthHeader = None
        
            
    
    if command == "createoauth":
        createOAuthToken(consumerKey, consumerSecret, SignatureMethod_RSA_SHA1())
        
    elif command == "listissues":
        listJiraIssues(project, auth, oAuthHeader, username, password)
        
    elif command == "createissue":
        createIssue(project, issueSummary, desc, issueType, priority, auth, oAuthHeader, username, password, assignee, parent)
    
    elif command == "initconfig":
        localVars = locals()
        configValues = {configKey : localVars[configKey] for configKey in cmdLineParams}
        initializeConfigFile(targetPath, None, **configValues)



