'''
Created on 21.05.2019

@author: sschaiter
'''

import requests
import logging
import os
import base64
import oauth2 as oauth
import configparser
from requests_oauthlib import OAuth1
from base64 import b64encode
from json.decoder import JSONDecodeError
from os.path import isfile
from Cryptodome.Signature import pkcs1_15
from Cryptodome.Hash import SHA1
from Cryptodome.PublicKey import RSA

log = logging.getLogger(__name__)


def createConfigFile(configPath : str = '', interactive = True):
    """
    Helper method to create a config file, and if necessary, missing directories on the path to it.
    
    :param configPath: The path to create an empty file on.
    :param interactive: Whether the CLI should ask for input regarding decisions. If this is False, 
    the assumption is made that user specified the correct path.
    :return: The path the config file was created on, or None if user declined to create it.
    """
    from clazzes.jiraclient.main import USER_CONFIG_FILEPATH, ETC_CONFIG_FILEPATH    
  
    if configPath and os.path.isfile(configPath):
        log.info("File already exists at path %s" %configPath)
        return configPath
            
    if interactive:
        gotValidAnswer = False    
        while not gotValidAnswer:  
            print("Enter absolute path to config file, starting with '/'.\n'u' will be substituted for %s\n'g' will be substituted for %s\nEnter 'n' to cancel.\n" \
              % (USER_CONFIG_FILEPATH, ETC_CONFIG_FILEPATH) )   
            configPath = input()
            if configPath:
                startLetter = configPath.lower()[0]
                if startLetter == 'u':
                    configPath = USER_CONFIG_FILEPATH
                    gotValidAnswer = True
                elif startLetter == 'g':
                    configPath = ETC_CONFIG_FILEPATH
                    gotValidAnswer = True
                elif startLetter == '/':
                    gotValidAnswer = True
                elif startLetter == 'n':
                    #Exit
                    return
                elif startLetter == '~':
                    configPath = os.path.expanduser(configPath)
                    gotValidAnswer = True
                else:
                    print("Invalid input. Please enter one of the given options.")
    
        if not os.path.isfile(configPath):
            
            gotValidAnswer = False
            while not gotValidAnswer:
                createNew = input("No file found on path %s, create it? (Y/n)\n(Note: JiraClient will attempt to create all directories on the path, if not present.)  " % configPath)
                if not createNew: #Empty input defaults to Yes
                    createNew = 'y'
                if createNew.lower() in ('y', 'n'):
                    gotValidAnswer = True
            if createNew == 'n':
                return
        #End of interactive block

    if not configPath:
        raise ValueError("Cannot create file at empty path")
    
    if os.path.isfile(configPath):
        log.info("File already exists at path %s" %configPath)
        return configPath
        
    try: #to catch IOErrors
        #Need to create directories up to the file name, if they don't exist yet
        dirsOnPath = configPath[1:].split('/')
        partPaths = []
        for step in dirsOnPath:
            partPaths.append('/' + '/'.join([substep for substep in dirsOnPath[:dirsOnPath.index(step)] ] ) )
        for path in partPaths:
            if not os.path.isdir(path):
                os.mkdir(path)
        #This actually creates the file
        fh = open(configPath, "w+")
        fh.close()
    except IOError as e:
        if e.errno == 13:
            raise ValueError("JiraClient does not have permissions to create the config file in path %s" %(configPath))
    return configPath

def makeApiCall(url : str, authMethod = 'OAUTH', oAuthHeader : OAuth1 = None, username : str = None, password : str = None, 
                params : dict = None, json : dict = None, method = "GET"):
    
    """
    Catch-all method to send Api Calls a Jira Server. Encapsulates Error Handling etc.
    
    :param url: The URL to make the call to.
    :param authMethod: The authentication method to be used.
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: The username to authenticate with. Only needed if using the "BASICAUTH" Authentication methd.
    :param password: The password to authenticate with. Only needed if using the "BASICAUTH" Authentication method.
    :param params: A dict to send in the request body. Mutually exclusive with json.
    :param json: A dict representing a JSON. Handled differently than, and mutually exclusive with, the params parameter. 
    :param method: The HTTP Method to be used. Throws an Exception if it doesn't exist. 
    """
    
    if params and json:
        raise ValueError("Both params and json were specified, which should never happen.")
    
    if authMethod == "BASICAUTH":
        b64authToken = b64encode(bytes("%s:%s" % (username, password), "UTF-8"))
        headers = {'Authorization' : "Basic " + str(b64authToken, "UTF-8")}
        auth = None
    elif authMethod == "OAUTH":
        headers = None
        auth = oAuthHeader
    else:
        raise ValueError("Submitted authorization type %s is not implemented" % authMethod)
    
    try:
        httpMethod = getattr(requests, method.lower())
    except:
        raise ValueError("HTTP method %s is not supported" % method)
    
    try:
        resp = httpMethod(url, params = params, json = json, auth = auth, headers = headers)
        resp.raise_for_status()
        respJson = resp.json()
    except requests.HTTPError:
        try: 
            raise requests.HTTPError("API Call to URL %s failed with status code %s and the following messages:\n%s %s" \
                                 % (url, resp.status_code, resp.json()['errorMessages'], resp.json()['errors']))
        except JSONDecodeError:
            raise requests.HTTPError("API Call to URL %s failed with status code %s" % (url, resp.status_code))
    except ValueError: #Thrown when the json decoding fails
        raise JSONDecodeError("Error decoding JSON received from URL %s with params %s" % (url, params) )
    return respJson

def getProjectIssueTypeIds(projectName : str, authMethod = 'OAUTH', oAuthHeader : OAuth1 = None, username : str = None, password : str = None):
    from clazzes.jiraclient.main import jiraApiUrl
    respJson = makeApiCall(url=jiraApiUrl + "issue/createmeta", oAuthHeader=oAuthHeader, username=username, 
                           password=password, params={'projectKeys' : projectName})
    if not respJson.get('projects', None):
        raise ValueError("Could not find project with name %s" % projectName)
    try:
        issueTypeJson = respJson.get('projects')[0].get('issuetypes')  
    except AttributeError:
        raise AttributeError('Did not receive a JSON response with expected format(containing project list with issue type list)')
    issueTypes = {issueType['name'] : issueType['id'] for issueType in issueTypeJson}
    return issueTypes

def sendJqlQuery(queryParams : dict, authMethod = 'OAUTH', oAuthHeader : OAuth1 = None, 
                 username : str = None, password : str = None):    
    """
    Fire off a search query that should return JSON descriptions of all objects meeting the criteria.
    Can search for Issues, projects, users etc.
    Type of object returned is inferred automatically by query params. For instance, "project=10" will search
    for projects with ID 10. See https://www.atlassian.com/blog/jira-software/jql-the-most-flexible-way-to-search-jira-14
    
    :param queryParams: dictionary with key-value pairs to search for.
    :param authMethod: The authentication method to be used.
    :param oAuthHeader: A :class:~requests.oauthlib.OAuth1` object which can be used for OAuth Authentication
    :param username: The username to authenticate with. Only needed if using the "BASICAUTH" Authentication methd.
    :param password: The password to authenticate with. Only needed if using the "BASICAUTH" Authentication method.
    """
    
    from clazzes.jiraclient.main import jiraApiUrl
    queryUrl = jiraApiUrl + "search"
    isFirst = True
    queryString = ""
    for param, value in queryParams.items():
        if not isFirst:
            queryString += "+AND+"
        queryString += "%s=%s" % (param, value)
        isFirst = False
    queryUrl = queryUrl + "?jql=" + queryString
    respJson = makeApiCall(queryUrl, authMethod, oAuthHeader, username, password)
    
    return respJson

def setConfigParams(cfgParams : dict, cfgPath : str, cfgParser = None):
    """
    Helper method to write one or more values to the [CLI] section of a given config(usually user or global config)
    If the path given does not point to a file, create a new file to write to.
    
    :param cfgParams: Dict containing the config values to be set.
    :param cfgPath: String containing the path to the config file. Preferably use path to user or global config file.
    :param cfgParser: To use an already created config parser. If None, create new one.
    """
    if cfgParser == None:
        cfgParser = makeConfigParser()
    #If the config file exists, read preset values
    if os.path.isfile(cfgPath): 
        cfgParser.read(cfgPath, "UTF-8")
    for name, value in cfgParams.items():
        cfgParser.set('CLI', name, value)
    if not isfile(cfgPath):
        #Create a new config file if it doesn't exist
        fh = open(cfgPath, 'w+')
        fh.close()
    with open(cfgPath, 'w') as cfgFile:
        cfgParser.write(cfgFile)
        
class SignatureMethod_RSA_SHA1(oauth.SignatureMethod):
    name = 'RSA-SHA1'

    def getSigningBase(self, request, consumer, token):
        """
        Assemble OAuth request string to be signed from OAuth Request Metadata.
        
        The Components used are the HTTP Method(POST, GET,...), the target URL, and various security parameters(normalized parameters)
        which include signing algorithm taken, the nonce, timestamp and OAuth Version. They are appended to the URL as a query string.
        
        :param request: A :class:~`oauth2.Request` with set 'method', 'normalized_url' attributes as well as several attributes
        needed for signature(returned by :meth:~`get_normalized_attributes` ).
        :param consumer: A :class:~`oauth2.Consumer` instance which is expected ot be initialized with both the shared secret and
        the label of its public key on the Resource Server.  
        :param token:
        
        :return: String containing the Method, URL and URL formatted query parameters. Special characters(?, =, / etc) are escaped.
        """
        if not hasattr(request, 'normalized_url') or request.normalized_url is None:
            raise ValueError("Base URL for request is not set.")

        sig = (
            oauth.escape(request.method),
            oauth.escape(request.normalized_url),
            oauth.escape(request.get_normalized_parameters()),
        )

        key = '%s&' % oauth.escape(consumer.secret)
        if token:
            key += oauth.escape(token.secret)
        raw = '&'.join(sig)
        return key, raw

    def sign(self, request, consumer, token):
        """
        Builds the request to be sent to an OAuth-supporting server, hashes it and signs it.
        
        :param key_location: Path to the file holding the private key to be used for signing. The key must correspond to the public
        key registered on Jira.
        :param request: A :class:~`oauth2.Request` with set 'method', 'normalized_url' attributes as well as several attributes
        needed for signature(returned by :meth:~`get_normalized_attributes` ).
        :param consumer: A :class:~`oauth2.Consumer` instance which is expected to be initialized with both the shared secret and
        the label of its public key on the Resource Server. 
        :param token: A :class:~`oauth2.Token` instance with both OAuthToken and OAuthTokenSecret attributes set.
        :return: A bytes object containing the b64-encoded hash of the assembled request string(see :meth:~`getSigningBase` )
        """
        _, raw = self.getSigningBase(request, consumer, token)
        
        from clazzes.jiraclient.main import privateKeyLocation
        with open(privateKeyLocation, 'r') as fh:
            data = fh.read()
        privateKeyString = data.strip()

        privatekey = RSA.import_key(privateKeyString)
        h = SHA1.new(raw.encode('UTF-8'))
        signature = pkcs1_15.new(privatekey).sign(h)

        return base64.b64encode(signature)
    
             
def makeConfigParser():
    """
    Create a fresh config parser instance.
    
    :return: A :class:~`ConfigParser` instance that has no file path set or values configured.
    """
    config = configparser.ConfigParser()
    config.optionxform = lambda option: option
    return config  